#!/usr/bin/env python
# coding: utf-8

# ##### Fashion-MNIST with tf.Keras, Margaret Maynard-Reid  
# https://medium.com/tensorflow/hello-deep-learning-fashion-mnist-with-keras-50fcff8cd74a

# Use CNNs on Fashion-MNIST dataset.

# In[2]:


from __future__ import absolute_import, division, print_function

# TensorFlow and tf.keras
import tensorflow as tf
from tensorflow import keras

# Helper libraries
import numpy as np
import matplotlib.pyplot as plt

print(tf.__version__)


# ## Import the Fashion MNIST dataset

# This guide uses the [Fashion MNIST](https://github.com/zalandoresearch/fashion-mnist) dataset which contains 70,000 grayscale images in 10 categories. The images show individual articles of clothing at low resolution (28 by 28 pixels), as seen here:
# 
# <table>
#   <tr><td>
#     <img src="https://tensorflow.org/images/fashion-mnist-sprite.png"
#          alt="Fashion MNIST sprite"  width="600">
#   </td></tr>
#   <tr><td align="center">
#     <b>Figure 1.</b> <a href="https://github.com/zalandoresearch/fashion-mnist">Fashion-MNIST samples</a> (by Zalando, MIT License).<br/>&nbsp;
#   </td></tr>
# </table>
# 
# Fashion MNIST is intended as a drop-in replacement for the classic [MNIST](http://yann.lecun.com/exdb/mnist/) dataset—often used as the "Hello, World" of machine learning programs for computer vision. The MNIST dataset contains images of handwritten digits (0, 1, 2, etc) in an identical format to the articles of clothing we'll use here.
# 
# This guide uses Fashion MNIST for variety, and because it's a slightly more challenging problem than regular MNIST. Both datasets are relatively small and are used to verify that an algorithm works as expected. They're good starting points to test and debug code. 
# 
# We will use 60,000 images to train the network and 10,000 images to evaluate how accurately the network learned to classify images. You can access the Fashion MNIST directly from TensorFlow, just import and load the data:

# In[11]:


fashion_mnist = keras.datasets.fashion_mnist

(x_train, y_train), (x_test, y_test) = fashion_mnist.load_data()


# Loading the dataset returns four NumPy arrays:
# 
# * The `x_train` and `y_train` arrays are the *training set*—the data the model uses to learn.
# * The model is tested against the *test set*, the `x_test`, and `y_test` arrays.
# 
# The images are 28x28 NumPy arrays, with pixel values ranging between 0 and 255. The *labels* are an array of integers, ranging from 0 to 9. These correspond to the *class* of clothing the image represents:
# 
# <table>
#   <tr>
#     <th>Label</th>
#     <th>Class</th> 
#   </tr>
#   <tr>
#     <td>0</td>
#     <td>T-shirt/top</td> 
#   </tr>
#   <tr>
#     <td>1</td>
#     <td>Trouser</td> 
#   </tr>
#     <tr>
#     <td>2</td>
#     <td>Pullover</td> 
#   </tr>
#     <tr>
#     <td>3</td>
#     <td>Dress</td> 
#   </tr>
#     <tr>
#     <td>4</td>
#     <td>Coat</td> 
#   </tr>
#     <tr>
#     <td>5</td>
#     <td>Sandal</td> 
#   </tr>
#     <tr>
#     <td>6</td>
#     <td>Shirt</td> 
#   </tr>
#     <tr>
#     <td>7</td>
#     <td>Sneaker</td> 
#   </tr>
#     <tr>
#     <td>8</td>
#     <td>Bag</td> 
#   </tr>
#     <tr>
#     <td>9</td>
#     <td>Ankle boot</td> 
#   </tr>
# </table>
# 
# Each image is mapped to a single label. Since the *class names* are not included with the dataset, store them here to use later when plotting the images:

# In[4]:


class_names = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat', 
               'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']


# ## Explore the data
# 
# Let's explore the format of the dataset before training the model. The following shows there are 60,000 images in the training set, with each image represented as 28 x 28 pixels.  Likewise, there are 60,000 labels in the training set.

# In[12]:


print("x_train shape:", x_train.shape, "y_train shape:", y_train.shape)


# Each label is an integer between 0 and 9:

# In[13]:


y_train


# There are 10,000 images in the test set. Again, each image is represented as 28 x 28 pixels. And the test set contains 10,000 images labels.

# In[15]:


x_test.shape, len(y_test)


# ## Visualize the data
# 
# Show one of the images from the training dataset.

# In[19]:


img_index=400
plt.imshow(x_train[img_index])


# ## Data normalization
# 
# Normalize the data dimensions so that they are of approximately the same scale.

# In[ ]:


x_train = x_train.astype('float32') / 255
x_test = x_test.astype('float32') / 255


# ## Data preprocessing
# 
# Reshape each image from (28,28) to (28,28,1), needed for convolutional layer

# In[34]:


x_train = x_train.reshape((60000,28,28,1))
x_test = x_test.reshape((10000,28,28,1))


# Convert `y` values from `int` to `categorical` (binary encoding)

# In[42]:


y_train = tf.keras.utils.to_categorical(y_train)
y_test = tf.keras.utils.to_categorical(y_test)


# In[47]:


print("x_train shape:", x_train.shape, "y_train shape:", y_train.shape)
print("x_test shape:", x_test.shape, "y_test shape:", y_test.shape)


# ## Split the data into train/validation/test datasets
# 
# Extract 20% of `x_train` and `y_train` as validation data.
# 
# - `test_size` is fraction extracted for validation
# - `random_state` is seed for random number generator

# In[43]:


from sklearn.model_selection import train_test_split
x_train_new, x_valid, y_train_new, y_valid = train_test_split(x_train, y_train, test_size=0.2, random_state=2018)


# In[46]:


print("x_train_new shape:", x_train_new.shape, "y_train_new shape:", y_train_new.shape)
print("x_valid shape:", x_valid.shape, "y_valid shape:", y_valid.shape)


# ## Build the model

# ### Setup the layers
# 
# There are two APIs for defining a model in Keras:
# 
# 1. Sequential model API
# 2. Functional API
# 
# In this tutorial we are using the Sequential model API to create a simple CNN model repeating a few layers of a convolution layer followed by a pooling layer then a dropout layer.
# 
# Note you only need to define the input data shape with the first layer. The last layers is a dense layer with softmax activation that classifies the 10 categories of data in fashion_mnist.

# In[20]:


model = tf.keras.Sequential()

# Must define the input shape in the first layer of the neural network
model.add(tf.keras.layers.Conv2D(filters=64, kernel_size=2, padding='same', activation='relu', input_shape=(28,28,1))) 
model.add(tf.keras.layers.MaxPooling2D(pool_size=2))
model.add(tf.keras.layers.Dropout(0.3))
model.add(tf.keras.layers.Conv2D(filters=32, kernel_size=2, padding='same', activation='relu'))
model.add(tf.keras.layers.MaxPooling2D(pool_size=2))
model.add(tf.keras.layers.Dropout(0.3))
model.add(tf.keras.layers.Flatten())
model.add(tf.keras.layers.Dense(256, activation='relu'))
model.add(tf.keras.layers.Dropout(0.5))
model.add(tf.keras.layers.Dense(10, activation='softmax'))

# Take a look at the model summary
model.summary()


# ### Compile the model
# 
# This is where you define the type of loss function, optimizer and the metrics evaluated by the model during training and testing.

# In[21]:


model.compile(loss='categorical_crossentropy',
             optimizer='adam',
             metrics=['accuracy'])


# ## Train the model
# 
# Training the neural network model requires the following steps:
# 
# 1. Feed the training data to the model—in this example, the `train_images` and `train_labels` arrays.
# 2. The model learns to associate images and labels.
# 3. We ask the model to make predictions about a test set—in this example, the `test_images` array. We verify that the predictions match the labels from the `test_labels` array. 
# 
# To start training,  call the `model.fit` method—the model is "fit" to the training data:

# In[ ]:


model.fit(x_train_new,
         y_train_new,
         batch_size=64,
         epochs=10,
         validation_data=(x_valid, y_valid)
         )


# As the model trains, the loss and accuracy metrics are displayed. This model reaches an accuracy of about 0.88 (or 88%) on the training data.

# ## Evaluate accuracy
# 
# Next, compare how the model performs on the test dataset:

# In[ ]:


test_loss, test_acc = model.evaluate(x_test, y_test)

print('Test accuracy:', test_acc)


# ## Make predictions
# 
# With the model trained, we can use it to make predictions about some images.

# In[ ]:


predictions = model.predict(x_test)


# Here, the model has predicted the label for each image in the testing set. Let's take a look at the first prediction:

# In[ ]:


predictions[0]


# A prediction is an array of 10 numbers. These describe the "confidence" of the model that the image corresponds to each of the 10 different articles of clothing. We can see which label has the highest confidence value:

# In[ ]:


np.argmax(predictions[0])


# So the model is most confident that this image is an ankle boot, or `class_names[9]`. And we can check the test label to see this is correct:

# In[ ]:


test_labels[0]


# We can graph this to look at the full set of 10 channels

# In[ ]:


def plot_image(i, predictions_array, true_label, img):
  predictions_array, true_label, img = predictions_array[i], true_label[i], img[i]
  plt.grid(False)
  plt.xticks([])
  plt.yticks([])
  
  plt.imshow(img, cmap=plt.cm.binary)

  predicted_label = np.argmax(predictions_array)
  if predicted_label == true_label:
    color = 'blue'
  else:
    color = 'red'
  
  plt.xlabel("{} {:2.0f}% ({})".format(class_names[predicted_label],
                                100*np.max(predictions_array),
                                class_names[true_label]),
                                color=color)

def plot_value_array(i, predictions_array, true_label):
  predictions_array, true_label = predictions_array[i], true_label[i]
  plt.grid(False)
  plt.xticks([])
  plt.yticks([])
  thisplot = plt.bar(range(10), predictions_array, color="#777777")
  plt.ylim([0, 1]) 
  predicted_label = np.argmax(predictions_array)
 
  thisplot[predicted_label].set_color('red')
  thisplot[true_label].set_color('blue')


# Let's look at the 0th image, predictions, and prediction array. 

# In[ ]:


i = 0
plt.figure(figsize=(6,3))
plt.subplot(1,2,1)
plot_image(i, predictions, test_labels, test_images)
plt.subplot(1,2,2)
plot_value_array(i, predictions,  test_labels)
plt.show()


# In[ ]:


i = 12
plt.figure(figsize=(6,3))
plt.subplot(1,2,1)
plot_image(i, predictions, test_labels, test_images)
plt.subplot(1,2,2)
plot_value_array(i, predictions,  test_labels)
plt.show()


# Let's plot several images with their predictions. Correct prediction labels are blue and incorrect prediction labels are red. The number gives the percent (out of 100) for the predicted label. Note that it can be wrong even when very confident. 

# In[ ]:


# Plot the first X test images, their predicted label, and the true label
# Color correct predictions in blue, incorrect predictions in red
num_rows = 5
num_cols = 3
num_images = num_rows*num_cols
plt.figure(figsize=(2*2*num_cols, 2*num_rows))
for i in range(num_images):
  plt.subplot(num_rows, 2*num_cols, 2*i+1)
  plot_image(i, predictions, test_labels, test_images)
  plt.subplot(num_rows, 2*num_cols, 2*i+2)
  plot_value_array(i, predictions, test_labels)
plt.show()


# Finally, use the trained model to make a prediction about a single image. 

# In[ ]:


# Grab an image from the test dataset
img = test_images[0]

print(img.shape)


# `tf.keras` models are optimized to make predictions on a *batch*, or collection, of examples at once. So even though we're using a single image, we need to add it to a list:

# In[ ]:


# Add the image to a batch where it's the only member.
img = (np.expand_dims(img,0))

print(img.shape)


# Now predict the image:

# In[ ]:


predictions_single = model.predict(img)

print(predictions_single)


# In[ ]:


plot_value_array(0, predictions_single, test_labels)
plt.xticks(range(10), class_names, rotation=45)
plt.show()


# `model.predict` returns a list of lists, one for each image in the batch of data. Grab the predictions for our (only) image in the batch:

# In[ ]:


prediction_result = np.argmax(predictions_single[0])
print(prediction_result)


# And, as before, the model predicts a label of 9.
